from os import write
from pathlib import Path
from . import Define as D
from . import preconts as K
from . import Pymaketool
from .module import AbstractModule, StaticLibraryModule
from . import Logger
from abc import ABC, abstractmethod

from pymakelib import module

log = Logger.getLogger()

def getLineSeparator(key: str, num: int):
    header = ''
    for _ in range(num):
        header += key
    return header

def macrosDictToString(macros):
    mstr = []
    if isinstance(macros, dict):
        for key in macros:
            if macros[key] != None and macros[key] != '':
                if isinstance(macros[key], str):
                    mstr.append('-D{}=\\\"{}\\\"'.format(key, macros[key]))
                elif isinstance(macros[key], bool):
                    mstr.append(
                        '-D{}={}'.format(key, '1' if macros[key] else '0'))
                elif isinstance(macros[key], D):
                    mstr.append(
                        '-D{}={}'.format(key, macros[key].getDefine()))
                else:
                    mstr.append('-D{}={}'.format(key, macros[key]))
            else:
                mstr.append('-D{}'.format(key))

    return ' '.join(mstr)


class Generator(ABC):

    def __init__(self, module: AbstractModule, project: Pymaketool):
        self.module = module
        self.project = project
        self.output = []

    @abstractmethod
    def process(self) -> str:
        pass

class MakeGenerator(Generator):

    def __init__(self, module, project: Pymaketool):
        super().__init__(module, project)
        self.isstaticlib = isinstance(module, StaticLibraryModule)
        if self.isstaticlib:
            mod:StaticLibraryModule = module
            mod.decorate_module()

    def write(self, value):
        self.output.append(value)

    def get_srcs_dirs(self, srcs) -> list:
        dirs = []
        for src in srcs:
            dirs.append(Path(str(src)).parent)
        dirs = list(set(dirs))
        return dirs

    def compiler_opts2str(self, compiler_opts):
        mstr = []
        def proc(moduleCompileOps):
            if isinstance(moduleCompileOps, dict):
                for key in moduleCompileOps:
                    if key == 'TARGETS':
                        continue
                    if (key == K.COMPOPTS_MACROS_KEY and isinstance(moduleCompileOps[key], dict)):
                        macros = macrosDictToString(moduleCompileOps[key])
                        mstr.append(macros)
                    else:
                        mstr.append(' '.join(moduleCompileOps[key]))

            elif isinstance(moduleCompileOps, list):
                for item in moduleCompileOps:
                    mstr.append(item)

        if isinstance(compiler_opts, dict):
            proc(compiler_opts)
        elif isinstance(compiler_opts, list):
            for moduleCompileOps in compiler_opts:
                proc(moduleCompileOps)

        rmstr = list(filter(lambda item: item, mstr))
        rmstr = ' '.join(rmstr)
        rmstr = ' '.join(rmstr.split())
        log.debug(f"compiler options: {rmstr}")
        return rmstr

    def write_header(self):
        mod_path =  f"{self.module.path}"
        self.write("{}\n".format(getLineSeparator('#', 52)))
        self.write("#{0:^50}#\n".format(mod_path))
        self.write("{}\n".format(getLineSeparator('#', 52)))
        self.write("\n")

    def write_srcs(self):
        prefixSrcs = ""
        if self.isstaticlib:
            prefixSrcs = self.module.name.upper() + "_"
        srcs = self.module.getSrcs()
        for src in srcs:
            if str(src).endswith('.c'):
                self.write("{}CSRC += {}\n".format(prefixSrcs, src))
            elif str(src).endswith('.cpp'):
                self.write("{}CXXSRC += {}\n".format(prefixSrcs, src))
            elif str(src).endswith('.s'):
                self.write("{}ASSRC += {}\n".format(prefixSrcs, src))
        self.write('\n') if srcs else None

    def write_incs(self):
        incs = self.module.getIncs()
        for inc in incs:
            if inc:
                self.write("INCS += -I{}\n".format(inc))
        self.write('\n') if incs else None

    def write_compiler_opts(self):
        proj_settings = self.project.projSettings;
        comp_opts = self.module.getCompilerOpts()
        if comp_opts:
            srcs = self.module.getSrcs()
            isstatic = self.isstaticlib
            for src in srcs:
                if isstatic:
                    objs = str(src).replace('.c', '.o').replace('.s', '.o')
                    outputObj = '$({}_OUTPUT)/'.format(self.module.key) + str(objs)
                    self.write("{} : CFLAGS = {}\n".format(str(outputObj), self.compiler_opts2str(comp_opts)))
                else:
                    objs = str(src).replace('.cpp', '.o')
                    objs = objs.replace('.c', '.o')
                    objs = objs.replace('.s', '.o')
                    ouputobj = Path(str(proj_settings['FOLDER_OUT']) + '/' + str(objs))
                    log.debug(ouputobj)
                    self.write("{} : CFLAGS = {}\n".format(str(ouputobj), self.compiler_opts2str(comp_opts)))


    def write_staticlib_def(self):
        mod: StaticLibraryModule = self.module
        log.debug(f"module \'{mod.path}\' static library name {mod.name}")
        log.debug(f"module \'{mod.path}\' static output directory {mod.output_dir}")
        self.write('{}_NAME = {}\n'.format(mod.key, mod.name))
        self.write('{}_OUTPUT = {}\n'.format(mod.key, str(mod.output_dir)))
        library = mod.output_dir / Path('lib'+mod.name+'.a')
        self.write('{}_AR = {}\n'.format(mod.key, str(library)))
        self.write('\n')

    def write_staticlib_rules(self):
        mod: StaticLibraryModule = self.module
        self.write('{0}'.format(mod.objects))
        self.write('\n\n' if mod.objects else '')
        self.write('{}'.format(mod.rule))
        self.write('\n\n' if mod.rule else '')
        self.write('{}\n'.format(mod.command))
        self.write('\n\n')
        self.write('SLIBS_NAMES += {}\n'.format(mod.linker))
        self.write('SLIBS_OBJECTS += {}\n'.format(mod.library))
        if mod.rebuild:
            self.write('\n')
            for src in mod.getSrcs():
                obj = str(src).replace('.c', '.o').replace('.s', '.o')
                outputObj = '$({}_OUTPUT)/'.format(mod.key) + str(obj)
                self.write("{} : .FORCE\n".format(outputObj))
            self.write('\n')

    def process(self) -> str:
        self.write_header()
        if self.isstaticlib:
            self.write_staticlib_def()
        self.write_srcs()
        self.write_incs()
        self.write_compiler_opts()
        if self.isstaticlib:
            self.write_staticlib_rules()
        return ''.join(self.output)



# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

PYMAKEPROJ  = '.pymakeproj'
MAKEFILE_PY = 'Makefile.py'
VARS_MK     = 'vars.mk'
TARGETS_MK  = 'targets.mk'
ECLIPSE_SETTING = '.settings'

MODULEIGNORE_FILE   = '.moduleignore'

MOD_WORKSPACE       = 'modPath'
MOD_COMPILER_OPTS   = 'compilerOpts'


MOD_F_INIT              = 'init'
MOD_F_GETSRCS           = 'getSrcs'
MOD_F_GETINCS           = 'getIncs'
MOD_F_GETCOMPILEROPTS   = 'getCompilerOpts'


MK_F_GETPROJECTSETTINGS = 'getProjectSettings'
MK_F_GETCOMPILERSET     = 'getCompilerSet'
MK_F_GETCOMPILEROPTS    = 'getCompilerOpts'
MK_F_GETLINKEROPTS      = 'getLinkerOpts'
MK_F_GETTARGETSSCRIPT   = 'getTargetsScript'


PROJSETT_PROJECTNAME    = 'PROJECT_NAME'
PROJSETT_FOLDEROUT      = 'FOLDER_OUT'


COMPILERSET_CC          = 'CC'
COMPILERSET_CXX         = 'CXX'
COMPILERSET_LD          = 'LD'
COMPILERSET_AR          = 'AR'
COMPILERSET_AS          = 'AS'
COMPILERSET_OBJCOPY     = 'OBJCOPY'
COMPILERSET_SIZE        = 'SIZE'
COMPILERSET_OBJDUMP     = 'OBJDUMP'
COMPILERSET_NM          = 'NM'
COMPILERSET_RANLIB      = 'RANLIB'
COMPILERSET_STRINGS     = 'STRINGS'
COMPILERSET_STRIP       = 'STRIP'
COMPILERSET_CXXFILT     = 'CXXFILT'
COMPILERSET_ADDR2LINE   = 'ADDR2LINE'
COMPILERSET_READELF     = 'READELF'
COMPILERSET_ELFEDIT     = 'ELFEDIT'


MK_KEY_MACROS               = 'MACROS'
MK_KEY_MACHINE_OPTS         = 'MACHINE-OPTS'
MK_KEY_OPTIMIZE_OPTS        = 'OPTIMIZE-OPTS'
MK_KEY_OPTIONS              = 'OPTIONS'
MK_KEY_DEBUGGING_OPTS       = 'DEBUGGING-OPTS'
MK_KEY_PREPROCESSOR_OPTS    = 'PREPROCESSOR-OPTS'
MK_KEY_WARNINGS_OPTS        = 'WARNINGS-OPTS'
MK_KEY_CONTROL_C_OPTS       = 'CONTROL-C-OPTS'
MK_KEY_GENERAL_OPTS         = 'GENERAL-OPTS'

COMPOPTS_MACROS_KEY = 'MACROS'



from pymakelib import module
import inspect
import logging
import sys

def __getmodule_caller(func):
    def stack_(frame):
        framelist = []
        while frame:
            framelist.append(frame)
            frame = frame.f_back
        return framelist

    stack_list = stack_(sys._getframe(1))
    idx = 0
    for stack in stack_list:
        idx += 1
        if stack.f_code.co_name == func:
            break
    parentframe = stack_list[idx]
    return parentframe.f_code.co_filename


def add_library(name, outputdir, srcs=[], incs=[]):

    @module.ModuleClass
    class _(module.BasicCModule, module.StaticLibraryModule):
                   
        def get_module_name(self):                  # 
            return str(name).capitalize()           #
                                                    # Not need in class mode
        def get_path(self):                         #
            return __getmodule_caller('add_library')  #

        def get_lib_name(self) -> str:
            return name

        def get_lib_outputdir(self) -> str:
            return outputdir

        def getSrcs(self):
            return super().getSrcs() if not srcs else srcs

        def getIncs(self):
            return super().getIncs() if not incs else incs


from .addon import AddonAbstract
from . import eclipse_cproject as cp

class EclipseAddon(AddonAbstract):
    """
    Generate Eclipse cproject files.
    """
    def init(self):
        self.generateLanguageSettings()
        self.generateCProject()

    def generateLanguageSettings(self):
        compSett = self.compilerSettings
        cp.generate_languageSettings(compSett)

    def generateCProject(self):
        projSett = self.projectSettings
        cp.generate_cproject(projSett)


# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# Generate .cproject from .cproject_template
#
# Wildcards:
#   <!--wildcard_c_includes-->      : C Includes
#   <!--wildcard_c_symbols-->       : C Symbols
#   <!--wildcard_cpp_includes-->    : C++ Includes
#   <!--wildcard_cpp_symbols-->     : C++ Symbols

# listconf = {'C_INCLUDES': [...], 'C_SYMBOLS' : [...], 'CPP_INCLUDES': [...], 'CPP_SYMBOLS': [...]}
from pathlib import Path
from . import preconts as K
from . import Define as D
import os.path
from os import path


CPROJECT_TEMPLATE = K.PYMAKEPROJ + '/.cproject_template'
CPROJECT = '.cproject'

LANGUAGE_SETTINGS_TEMPLATE = K.PYMAKEPROJ + '/.language.settings_template'
LANGUAGE_SETTINGS = '.settings/language.settings.xml'

LANGUAGE_SETTING_PROVIDER = """
                        <provider class="org.eclipse.cdt.managedbuilder.language.settings.providers.GCCBuiltinSpecsDetector" console="false" env-hash="-374963487674273350" id="org.eclipse.cdt.managedbuilder.core.GCCBuiltinSpecsDetector" keep-relative-paths="false" name="CDT GCC Built-in Compiler Settings" parameter="{0} -E -P -v -dD &quot;${{INPUTS}}&quot;" prefer-non-shared="true">
                                <language-scope id="org.eclipse.cdt.core.gcc"/>
                                <language-scope id="org.eclipse.cdt.core.g++"/>
                        </provider>
"""

WILDCARD_C_INCLUDES = '<!--wildcard_c_includes-->'
WILDCARD_C_SYMBOLS = '<!--wildcard_c_symbols-->'
WILDCARD_C_EXCLUDE = '<!--wildcard_c_exclude-->'

WILDCARD_LS_PROVIDER = '<!--wildcard_ls_provider-->'

def generate_cproject(listconf: dict):
    print('Generate .cproject')
    try:
        cproject_template = open(CPROJECT_TEMPLATE, 'r')
        cproject = open(CPROJECT, 'w')

        for line in cproject_template:
            if(line.strip() == WILDCARD_C_INCLUDES):
                if listconf['C_INCLUDES']:
                    cproject.write((writeXmlIncludes(listconf['C_INCLUDES'])))
            if(line.strip() == WILDCARD_C_SYMBOLS):
                if listconf['C_SYMBOLS']:
                    cproject.write((writeXmlSymbols(listconf['C_SYMBOLS'])))
            if(line.strip() == WILDCARD_C_EXCLUDE):
                if listconf['C_EXCLUDE']:
                    cproject.write(writeXmlExcluding(listconf['C_EXCLUDE']))    
            else:
                cproject.write(line)

    except IOError:
        print('Files .cproject or .cproject_template no accessible')
    finally:
        cproject_template.close()
        cproject.close()

def generate_languageSettings(compilerSettings: dict):
    print('Generate .setting/language.settings.xml')
    try:
        try:
            if not (path.exists(K.ECLIPSE_SETTING)):
                os.mkdir(K.ECLIPSE_SETTING)
        except Exception as e:
            print(e)
            
        langsett_template = open(LANGUAGE_SETTINGS_TEMPLATE, 'r')
        langsett = open(LANGUAGE_SETTINGS, 'w')

        for line in langsett_template:
            if(line.strip() == WILDCARD_LS_PROVIDER):
                langsett.write(LANGUAGE_SETTING_PROVIDER.format(compilerSettings['CC']))
            else:
                langsett.write(line)

    except IOError:
        print('Files .cproject or .cproject_template no accessible')
    finally:
        langsett_template.close()
        langsett.close()


def writeXmlIncludes(incList):
    directory = Path("./srcs.mk")
    directory = str(directory.absolute().parent.name)
    w = []
    for i in incList:
        if str(i).startswith('/'):  # absolute path
            w.append("<listOptionValue builtIn=\"false\" value=\"" +
                     str(i) + "\"/>\n")
        else:  # realative path
            w.append(
                "<listOptionValue builtIn=\"false\" value=\"&quot;${workspace_loc:/"+directory+"/"+str(i)+"}&quot;\"/>\n")

    return ''.join(w)


def writeXmlExcluding(excList):
    w = [] 
    excludes = '|'.join(excList)
    w.append("<entry excluding=\"" + excludes + "\" flags=\"VALUE_WORKSPACE_PATH\" kind=\"sourcePath\" name=\"\"/>\n")
    return ''.join(w)

def writeXmlSymbols(symList):
    w = []
    
    if isinstance(symList, dict):
        for key in symList:
            if symList[key] != None and symList[key] != '':
                if isinstance(symList[key], str):
                    w.append('<listOptionValue builtIn=\"false\" value=\"{}=&quot;{}&quot;\"/>\n'.format(key, symList[key]))
                elif isinstance(symList[key], bool):
                    w.append('<listOptionValue builtIn=\"false\" value=\"{}={}\"/>\n'.format(key, '1' if symList[key] else '0'))
                elif isinstance(symList[key],D):
                    w.append('<listOptionValue builtIn=\"false\" value=\"{}={}\"/>\n'.format(key, symList[key].getDefine()))
                else:
                    w.append('<listOptionValue builtIn=\"false\" value=\"{}={}\"/>\n'.format(key, symList[key]))
            else:
                w.append('<listOptionValue builtIn=\"false\" value=\"{}\"/>\n'.format(key))
    else:
        for sym in symList:
            sym = str(sym).replace("\\\"", "&quot;")
            w.append("<listOptionValue builtIn=\"false\" value=\""+sym+"\"/>\n")

    return ''.join(w)



from os import execle
import tarfile
import tempfile
import re
from pathlib import Path
from abc import ABC,abstractmethod
from prompt_toolkit import prompt
from prompt_toolkit.completion import PathCompleter, WordCompleter
from prompt_toolkit.validation import Validator
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from . import Logger

log = Logger.getLogger()

class Object(object):
    pass

class PromptUtil:

    def __init__(self) -> None:
        self.bindings = KeyBindings()
        
        @self.bindings.add(Keys.ControlC)
        def _(event):
            exit(0)

    def add_answers(self, answers: dict):
        self.answers = answers

    def parse_answers(self):
        obj = Object()
        for a in self.answers:
            type = a['type']
            name = a['name']
            self.answer = a
            res = getattr(self, type)()
            setattr(obj, name, res)
        return obj


    def __get_commond_attr(self):
        msg = self.answer['msg']
        default = '' if not "default" in self.answer else self.answer["default"]
        validator_dict = None if not "validator" in self.answer else self.answer["validator"]
        validator = None
        if validator_dict:
            validator = Validator.from_callable(
                validator_dict['callback'],
                error_message=validator_dict['error_msg'],
                move_cursor_to_end=True)
        return msg, default, validator


    def input(self):
        msg, default, validator = self.__get_commond_attr()
        return prompt(msg, default=default, validator=validator, key_bindings=self.bindings)


    def input_path(self):
        msg, default, validator = self.__get_commond_attr()
        path_completer = PathCompleter(only_directories=True, min_input_len=1)
        return prompt(msg, completer=path_completer, default=default, validator=validator, key_bindings=self.bindings)


    def confirm(self):
        msg, default, validator = self.__get_commond_attr()
        word_completer = WordCompleter(["yes", "no"])
        resp = prompt(msg, completer=word_completer, default=default, validator=validator, key_bindings=self.bindings)
        return True if resp == 'yes' or resp == 'y' else False



class AbstractGenerator(ABC):

    @abstractmethod
    def info(self) -> dict:
        pass

    @abstractmethod
    def exec_generator(self, **kwargs):
        pass


class BasicGenerator(AbstractGenerator):

    @abstractmethod
    def get_attrs(self, **kwargs) -> dict:
        pass


    def parse_args(self, args:list) -> dict:
        res = []
        for sub in args:
            if '=' in sub:
                res.append(map(str.strip, sub.split('=', 1)))
            else:
                res.append((sub, None))
        res = dict(res)
        return res


    def copyFile(self, input_file, output_file, tokens: dict):
        p = Path(output_file.parent)
        p.mkdir(parents=True, exist_ok=True)
        input_file = open(input_file, 'r')
        output_file = open(output_file, 'w')
        for line in input_file:
            for token, value in tokens.items():
                line = re.sub("<%=[ ]*" + token + "[ ]*%>", value, line)
            output_file.write(line)

    def exec_generator(self, **kwargs):
        try:
            attrs = self.get_attrs(**kwargs)
            self.temp_name      = attrs['temp_name']
            self.temp_gzip_file = attrs['temp_gzip_file']
            self.temp_files     = attrs['temp_files']
            self.temp_tokens    = attrs['temp_tokens']
            self.output_folder  = attrs['output_folder']

            output_folder = Path(self.output_folder)
            output_folder.mkdir()

            gzipfile = Path(self.temp_gzip_file)
            if not  gzipfile.exists():
                print(f"File {str(gzipfile)} not found")
                exit(1)

            tmpdir = tempfile.TemporaryDirectory()
            gzipobj = tarfile.open(gzipfile)
            gzipobj.extractall(tmpdir.name)
            gzipobj.close()

            dir_tmptemp = Path(Path(tmpdir.name) / Path(self.temp_name))

            for tp in self.temp_files:
                fin = Path(dir_tmptemp / Path(tp))
                if fin.exists():
                    fout = Path( output_folder / Path(tp))
                    log.debug("{0} > {1}".format(str(fin), str(fout)))
                    if fin.is_dir():
                        fout.mkdir(parents=True, exist_ok=True)
                    else:
                        self.copyFile(fin, fout, self.temp_tokens)
        except FileExistsError:
            print(f"Folder {output_folder} already exist")
        except Exception as ex:
            log.exception(ex)



import inspect

initAddonFuncs = []
initAddonClass = []

def init(func):
    initAddonFuncs.append(func)


def add(inst):
    if inspect.isclass(inst):
        initAddonClass.append(inst)
    elif inspect.isfunction(inst):
        initAddonFuncs.append(inst)


class AddonAbstract:
    def __init__(self, projectSettings, compilerSettings):
        self.projectSettings = projectSettings
        self.compilerSettings = compilerSettings
    def init(self):
        pass


import subprocess
import tempfile
from pathlib import Path

def getGCCHeaderFiles(cmd_gcc):
    gcc_includes = []
    try:
        auxfile = tempfile.NamedTemporaryFile()
        command = "echo | {0} -Wp,-v -x c++ - -fsyntax-only &> {1} ; cat {1} |  grep -e '^\s.*'".format(
            cmd_gcc, auxfile.name)
        res = subprocess.check_output(["bash", "-c", command])
        for line in res.splitlines():
            gcc_includes.append(line.decode('utf-8').strip())
        auxfile.close()
    except Exception as e:
        print(e)
    return gcc_includes


def get_gcc_arm_none_eabi(binLocation='', prefix='arm-none-eabi-', extIncludes=[]) -> dict:
    """Get dictionary with gcc compiler set of arm-none-eabi-

    Args:
        binLocation (str, optional): Location of toolchain binary. Defaults to ''.
        prefix (str, optional): prefix of ARM toolchain. Defaults to 'arm-none-eabi-'.
        extIncludes (list, optional): list of external includes. Defaults to [].

    Returns:
        dict: set of gcc compiler e.g. {'CC': 'arm-none-eabi-gcc' ... }
    """    
    return confGCC(binLocation, prefix, extIncludes)


def confARMeabiGCC(binLocation='', prefix='arm-none-eabi-', extIncludes=[]):
    return confGCC(binLocation, prefix, extIncludes)


def get_c_linux(bin_location='', ext_incs=[]) -> dict:
    return get_gcc_linux(bin_location, ext_incs)


def get_gcc_linux(bin_location='', ext_incs=[]) -> dict:
    """Get dictionary with gcc compiler set for linux   

    Args:
        bin_location (str, optional): location of toolchain. Defaults to ''.
        ext_incs (list, optional): list of external includes. Defaults to [].

    Returns:
        dict: set of gcc compiler e.g. {'CC': 'gcc' ... }
    """    
    return confLinuxGCC(binLocation=bin_location, extIncludes=ext_incs)


def get_cpp_linux(bin_location='', ext_incs=[]) -> dict:
    return get_gpp_linux(bin_location, ext_incs)


def get_gpp_linux(bin_location='', ext_incs=[]) -> dict:
    """Get dictionary with g++ compiler set for linux   

    Args:
        bin_location (str, optional): location of toolchain. Defaults to ''.
        ext_incs (list, optional): list of external includes. Defaults to [].

    Returns:
        dict: set of gcc compiler e.g. {'CC': 'g++' ... }
    """    
    return confGCC(binLocation=bin_location, extIncludes=ext_incs, iscpp=True)


def confLinuxGCC(binLocation='', extIncludes=[]):
    return confGCC(binLocation, '', extIncludes)


def confGCC(binLocation='', prefix='', extIncludes=[], iscpp=False):
    binpath = Path(binLocation)
    cmd_gcc = str(binpath / (prefix + 'gcc'))
    cmd_gxx = str(binpath / (prefix + 'g++'))
    cmd_ld = str(binpath / (prefix + ('g++' if iscpp else 'gcc')))
    cmd_ar = str(binpath / (prefix + 'ar'))
    cmd_as = str(binpath / (prefix + 'as'))
    cmd_objcopy = str(binpath / (prefix + 'objcopy'))
    cmd_size = str(binpath / (prefix + 'size'))
    cmd_objdump = str(binpath / (prefix + 'objdump'))
    cmd_nm = str(binpath / (prefix + 'nm'))
    cmd_ranlib = str(binpath / (prefix + 'ranlib'))
    cmd_strings = str(binpath / (prefix + 'strings'))
    cmd_strip = str(binpath / (prefix + 'strip'))
    cmd_cxxfilt = str(binpath / (prefix + 'c++filt'))
    cmd_addr2line = str(binpath / (prefix + 'addr2line'))
    cmd_readelf = str(binpath / (prefix + 'readelf'))
    cmd_elfedit = str(binpath / (prefix + 'elfedit'))

    gcc_includes = getGCCHeaderFiles(cmd_gcc)
    gcc_includes = gcc_includes + extIncludes
    return confToolchain(cmd_gcc,
                         cmd_gxx,
                         cmd_ld,
                         cmd_ar,
                         cmd_as,
                         cmd_objcopy,
                         cmd_size,
                         cmd_objdump,
                         cmd_nm,
                         cmd_ranlib,
                         cmd_strings,
                         cmd_strip,
                         cmd_cxxfilt,
                         cmd_addr2line,
                         cmd_readelf,
                         cmd_elfedit,
                         gcc_includes)


def confToolchain(
    cmd_gcc,
    cmd_gxx,
    cmd_ld,
    cmd_ar,
    cmd_as,
    cmd_objcopy,
    cmd_size,
    cmd_objdump,
    cmd_nm,
    cmd_ranlib,
    cmd_strings,
    cmd_strip,
    cmd_cxxfilt,
    cmd_addr2line,
    cmd_readelf,
    cmd_elfedit,
    includes
):
    return {
        'CC':       cmd_gcc,
        'CXX':      cmd_gxx,
        'LD':       cmd_ld,
        'AR':       cmd_ar,
        'AS':       cmd_as,
        'OBJCOPY':  cmd_objcopy,
        'SIZE':     cmd_size,
        'OBJDUMP':  cmd_objdump,
        'NM':       cmd_nm,
        'RANLIB':   cmd_ranlib,
        'STRINGS':  cmd_strings,
        'STRIP':    cmd_strip,
        'CXXFILT':  cmd_cxxfilt,
        'ADDR2LINE':cmd_addr2line,
        'READELF':  cmd_readelf,
        'ELFEDIT':  cmd_elfedit,
        'INCLUDES': includes
    }



# Copyright (c) 2020, Ericson Joseph
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import hashlib
import re
import importlib.util
import copy
from pathlib import Path
from . import preconts as K
from . import git
from abc import ABC,abstractmethod, ABCMeta
from . import Logger

log = Logger.getLogger()

class SrcType:
    C = ['.c']
    CPP = ['.C', '.cc', '.cpp', '.CPP', '.c++', '.cp', '.cxx']
    ASM = ['.s', '.S', '.asm']


class IncType:
    C = ['.h']
    CPP = ['.h', '.hpp', '.h++', '.hh']


class StaticLibrary:
    def __init__(self, name: str, outputDir: str, rebuild=False, lib_linked_opts=None, orden=1):
        self.name = name
        self.orden = orden
        self.outputDir = Path(outputDir)
        self.rebuild = rebuild
        self.mkkey = self.name.upper()
        self.lib_name = 'lib' + self.name + '.a'
        self.library = self.outputDir / Path(self.lib_name)
        self.lib_objs = f"{self.mkkey}_OBJECTS = $({self.mkkey}_CSRC:%.c=$({self.mkkey}_OUTPUT)/%.o) $({self.mkkey}_CSRC:%.s=$({self.mkkey}_OUTPUT)/%.o)"
        self.lib_objs_compile = f"$({self.mkkey}_OUTPUT)/%.o: %.c\n\t$(call logger-compile-lib,\"CC\",\"{self.library}\",$<)\n\t@mkdir -p $(dir $@)\n\t$(CC) $(CFLAGS) $(INCS) -o $@ -c $<"
        self.lib_compile = f"$({self.mkkey}_AR): $({self.mkkey}_OBJECTS)\n\t$(call logger-compile,\"AR\",$@)\n\t$(AR) -rc $@ $(filter %.o,$({self.mkkey}_OBJECTS))"
        self.lib_linked_opts = lib_linked_opts
        self.lib_linked = "-L{0} -l{1} {2}".format(self.outputDir, self.name, self._get_str_linked_opts(self.lib_linked_opts))

    def _get_str_linked_opts(self, opts):
        if isinstance(opts, str):
            return opts
        elif isinstance(opts, list):
            return ' '.join(opts)
        else:
            return ""

    def setRebuild(self, rebuild: bool):
        self.rebuild = rebuild

    def rebuildByCheckStr(self, checkStr: str):
        hs = hashlib.md5(checkStr.encode())
        cksumfile = Path(self.outputDir / Path('lib' + self.name + '.cksum'))
        hexstr = hs.hexdigest()
        oldHash = ""
        if cksumfile.exists():
            cksumfile = open(str(cksumfile), 'r+')
            oldHash = cksumfile.read()
        else:
            cksumfile = open(str(cksumfile), 'w')
        
        if hexstr != oldHash:
            cksumfile.seek(0)
            cksumfile.truncate(0)
            cksumfile.seek(0)
            cksumfile.write(str(hexstr))
            self.setRebuild(True)

        cksumfile.close()

class Module:
    def __init__(self, srcs, incs, flags, filename, staticLib: StaticLibrary = None):
        self.srcs = srcs
        self.incs = incs
        self.flags = flags
        self.filename = filename
        self.module_name = ""
        self.staticLib = staticLib
        self.orden = 0 if staticLib == None else staticLib.orden
    
    def isEmpty(self):
        if not self.srcs and not self.incs and not self.staticLib:
            return True
        else:
            return False

    def getDirs(self):
        dirs = []
        for src in self.srcs:
            dirs.append(Path(str(src)).parent)
        dirs = list(set(dirs))
        return dirs

class CompilerOptions:
    def __init__(self, opts: dict):
        self.opts = opts

    def setOption(self, key, value):
        self.opts[key] = value
        return self.opts

    def addOption(self, key, value):
        if key in self.opts.keys():
            if isinstance(value, str):
                self.opts[key].append(value)
            elif isinstance(value, list):
                self.opts[key] = self.opts[key] + value
        else:
            self.opts[key] = value

class GCC_CompilerOpts(CompilerOptions):
    def __init__(self, copts):
        if isinstance(copts, CompilerOptions):
            super().__init__(copts.opts)
        elif isinstance(copts, dict):
            super().__init__(copts)
        else:
            pass

    def addGeneralOpt(self, opts:list):
        self.addOption(K.MK_KEY_GENERAL_OPTS, opts)

    def setOptimizationOpts(self, opts: list):
        self.setOption(K.MK_KEY_OPTIMIZE_OPTS, opts)

    def setControlCOpts(self, opts: list):
        self.setOption(K.MK_KEY_CONTROL_C_OPTS, opts)

    def addMacroOpts(self, macro, value=None):
        self.opts[K.MK_KEY_MACROS][macro] = value

    def getMacroValue(self, macro):
        return self.opts[K.MK_KEY_MACROS][macro]

    def setDebuggingOpts(self, opts: list):
        self.setOption(K.MK_KEY_DEBUGGING_OPTS, opts)

    def setWarningdOpts(self, opts: list):
        self.setOption(K.MK_KEY_WARNINGS_OPTS, opts)

    def isDefine(self, macro):
        if macro in self.opts[K.MK_KEY_MACROS].keys():
            return True
        else:
            return False

    def isMacroValue(self, macro, value):
        if self.isDefine(macro):
            return True if self.getMacroValue(macro) == value else False
        else:
            return False


class ModuleHandle:
    def __init__(self, modDir, gCompOpts, goal=None):
        self.modDir = modDir
        self.gCompOpts = CompilerOptions(gCompOpts)
        self.goal = goal

    def getWorkspace(self):
        wk = {
            K.MOD_WORKSPACE: self.modDir,
            K.MOD_COMPILER_OPTS: self.gCompOpts.opts
        }
        return wk

    def getAllSrcsC(self):
        return self.getAllSrcs(SrcType.C)

    def getAllSrcs(self, srcType: SrcType):
        srcs = []
        for ext in srcType:
            srcs += list(Path(self.modDir).rglob('*' + ext))
        return srcs

    def getFilesByRegex(self, regexs, relativePath=None):
        modulePath = Path(self.modDir)
        if relativePath:
            modulePath = modulePath / Path(relativePath)
        
        srcs = []
        for r in regexs:
            srcs += list(modulePath.rglob(r))

        srcs = list(dict.fromkeys(srcs))
        return srcs

    def getAllIncs(self, incType: IncType):
        incsfiles = []
        for ext in incType:
            incsfiles += list(Path(self.modDir).rglob('*' + ext))

        incs = []
        for i in incsfiles:
            incs.append(i.parent)

        incs = list(dict.fromkeys(incs))
        return incs

    def getAllIncsC(self):
        return self.getAllIncs(IncType.C)

    def getFileByNames(self, names):
        return self.getFilesByRegex(names)

    def getSrcsByPath(self, srcs):
        srcslist = []
        for src in srcs:
            srcslist.append(Path(Path(self.modDir) / src))
        return srcslist

    def getGeneralCompilerOpts(self):
        return self.gCompOpts

    def getRelaptivePath(self):
        return self.modDir

    @DeprecationWarning
    def initGitModule(self, url, folder, ispymakeproj=False, ignoreList=[]):
        absfolder = Path(Path(self.getRelaptivePath()) / Path(folder))
        
        ignoreModule = []
        ignoreModule += self.getFilesByRegex(ignoreList, relativePath=Path(folder))
        
        git.addSubmodule(url, str(absfolder), ispymakeproj, ignoreModule)

    def getGoal(self):
        return self.goal

    def __str__(self):
        return str(self.modDir) + " " + str(self.gCompOpts)

class AbstractModule(ABC):
    """Abstract class of pymaketool module

    Args:
        path (str): path to module, _mk.py file.

    Attributes:
        path (str): path of module
    """    
    def __init__(self) -> None:
        super().__init__()
        self.module_name = self.get_module_name()
        self.path = self.get_path()
        self.dir = Path(self.path).parent

    def get_module_name(self) -> str:
        """Module name

        Returns:
            str: name of module (default: class name)
        """
        return self.__class__.__name__

    def get_path(self):
        """Get path of module in filesystem

        Returns:
            str: path (default: self.__module__.module_name)
        """
        return f"{self.__module__}.{self.module_name}"

    def init(self):
        """Initialization of module
        """        
        pass

    @abstractmethod
    def getSrcs(self) -> list:
        """Abstract method to get the sources paths of module

        Returns:
            list: list of sources paths realtive to project
        """        
        pass
    
    @abstractmethod
    def getIncs(self) -> list:
        """Abstract method to get the includes paths of module

        Returns:
            list: list of includes paths realtive to project
        """        
        pass
    
    def findSrcs(self, src_type: SrcType) -> list:
        """Util method for find sources in module path

        Args:
            src_type (SrcType): Type of sources C, CPP or ASM

        Returns:
            list: list of sources paths realtive to project
        """        
        log.debug(f"find srcs in {self.dir}")
        srcs = []
        for ext in src_type:
            srcs += list(Path(self.dir).rglob('*' + ext))
        return srcs
        
    def findIncs(self, inc_type: IncType) -> list:
        """Util method for find includes in module path

        Args:
            inc_type (IncType): Type of includes C or CPP

        Returns:
            list: list of includes paths realtive to project
        """        
        incsfiles = []
        for ext in inc_type:
            incsfiles += list(Path(self.dir).rglob('*' + ext))

        incs = []
        for i in incsfiles:
            incs.append(i.parent)

        incs = list(dict.fromkeys(incs))
        return incs

    def getAllSrcsC(self) -> list:
        """Util method for get all sources in module, type C

        Returns:
            list: list of sources paths realtive to project
        """        
        return self.findSrcs(SrcType.C)

    def getAllIncsC(self) -> list:
        """Ütil method for get all includes in module, type C 

        Returns:
            list: list of includes paths realtive to project
        """        
        return self.findIncs(IncType.C)

    def getCompilerOpts(self):
        """Get special compiler options for module
        """        
        pass

class StaticLibraryModule(metaclass=ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_lib_name') and 
                callable(subclass.get_lib_name) and
                hasattr(subclass, 'get_lib_outputdir') and 
                callable(subclass.get_lib_outputdir))

    def decorate_module(self):
        self.name = self.get_lib_name()
        self.lib_name = 'lib' + self.name + '.a'
        self.output_dir = Path(self.get_lib_outputdir())
        self.orden = self.get_order()
        self.rebuild = self.get_rebuild()
        self.key = self.name.upper()
        self.library = self.output_dir / Path(self.lib_name)
        self.objects = self.get_objects(self.key)
        self.rule = self.get_rule(self.key)
        self.command = self.get_command(self.key)
        self.linker = self.get_linker(self.key)

    def _get_str_linked_opts(self, opts):
        if isinstance(opts, str):
            return opts
        elif isinstance(opts, list):
            return ' '.join(opts)
        else:
            return ""

    @abstractmethod
    def get_lib_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_lib_outputdir(self) -> str:
        raise NotImplementedError

    def get_linker_opts(self) -> str:
        return None

    def get_objects(self, key) -> str:
        return  f"{key}_OBJECTS = $({key}_CSRC:%.c=$({key}_OUTPUT)/%.o) $({key}_CSRC:%.s=$({key}_OUTPUT)/%.o)"

    def get_rule(self, key) -> str:
        return  f"$({key}_OUTPUT)/%.o: %.c\n\t$(call logger-compile-lib,\"CC\",\"{key}\",$<)\n\t@mkdir -p $(dir $@)\n\t$(CC) $(CFLAGS) $(INCS) -o $@ -c $<"

    def get_command(self, key) -> str:
        return  f"$({key}_AR): $({key}_OBJECTS)\n\t$(call logger-compile,\"AR\",$@)\n\t$(AR) -rc $@ $(filter %.o,$({key}_OBJECTS))"

    def get_linker(self, key) -> str:
        return "-L{0} -l{1}".format(self.output_dir, self.name, self._get_str_linked_opts(self.get_linker_opts()))

    def get_order(self):
        return 1

    def get_rebuild(self):
        return False


class POJOModule(AbstractModule):
    def __init__(self, path):
        super().__init__(path)
        self.includes = []
        self.sources = []
        self.compiler_opts = None
        self.init_resp = None

    def init(self):
        return self.init_resp

    def getSrcs(self) -> list:
        return self.includes

    def getIncs(self) -> list:
        return self.sources

    def getCompilerOpts(self):
        return self.compiler_opts   

class BasicCModule(AbstractModule):
    """Basic C module, find all sources and includes in module path

    Args:
        path (str): path to module, _mk.py file.
    """
    def __init__(self):
        super().__init__()

    def getSrcs(self) -> list:
        """Return list with all sources in module path

        Returns:
            list: sources paths
        """        
        return self.getAllSrcsC()

    def getIncs(self) -> list:
        """return list with all includes in module path

        Returns:
            list: includes path 
        """        
        return self.getAllIncsC()

class ExternalModule(AbstractModule):
    """The ExternalModule object that inherits from AbstractModule for include external pymaketool module 

    Args:
        path (str): path to module, _mk.py file.

    Attributes:
        remoteModule (AbstractModule): remote module object.
    
    Raises:
            AttributeError: path is not valid
    """
    def __init__(self):
        super().__init__()
        try:
            modPath = self.getModulePath()
            if not modPath.endswith("_mk.py"):
                raise AttributeError(f"{modPath} is not a valid module path")
            lib = importlib.util.spec_from_file_location(str(modPath), str(modPath))
            mod = importlib.util.module_from_spec(lib)
            lib.loader.exec_module(mod)
            obj = getModuleInstance()[0]
            log.debug(f"create copy of module object {obj.__class__}")
            self.remoteModule = copy.deepcopy(obj)
            cleanModuleInstance()
        except Exception as ex:
            log.exception(ex)
            exit(-1)
    
    @abstractmethod
    def getModulePath(self)->str:
        """Abstract methos to get string path of external module

        Returns:
            str: path of external module
        """
        pass

    def init(self):
        """Call and return init from remote module

        Returns:
            object: may be StaticLibrary object or None
        """
        try:
            return self.remoteModule.init()
        except AttributeError as ae:
            log.debug(ae)
        except Exception as ex:
            log.exception(ex)
            exit(-1)

    def getSrcs(self):
        """Call and return getSrcs from remote module

        Returns:
            list: list of sources
        """
        return self.remoteModule.getSrcs()
        
    def getIncs(self):
        """Call and return getIncs from remote module

        Returns:
            list: list of includes
        """
        return self.remoteModule.getIncs()
    
    def getCompilerOpts(self):
        """Call and return getCompilerOpts from remote module

        Returns:
            disct: compiler options
        """
        try:
            return self.remoteModule.getCompilerOpts()
        except AttributeError as ae:
            log.debug(ae)
        except Exception as ex:
            log.exception(ex)
            exit(-1)


def ModuleClass(clazz):
    """Add class to modules of pymaketool

    Args:
        clazz (class): Class inheritance of Module.AbstractModule
    """    
    if issubclass(clazz, AbstractModule):
        log.debug(f"class \'{clazz.__name__}\' is inheritance of Module.AbstractModule")
    else:
        log.warning(f"class \'{clazz.__name__}\' in \'{__name__}\' not inheritance of Module.AbstractModule")

    # classdir = str(clazz)
    # log.debug(f"class dir {classdir}")
    # m = re.search(r"<class \'(?P<dir>[a-zA-Z\./_-]+)\'>", classdir)
    # modulePath = None
    # if m:
    #     p = Path(m.group('dir'))
    #     modulePath = p
    # modulePath = f"{clazz.__module__}.{clazz.__name__}"

    obj = clazz()
    global ModulesInstances
    try:
        _ = ModulesInstances
    except NameError:
        log.debug(f"create global modules list")
        ModulesInstances = []

    
    log.debug(f"add new instance of ModuleClass \'{clazz.__name__}\' with path {clazz.__module__}")
    ModulesInstances.append(obj)


def getModuleInstance() -> AbstractModule:
    try:
        _ = ModulesInstances
        return ModulesInstances
    except NameError:
        log.debug("not ModuleClass in project")
        pass
    return None


def cleanModuleInstance():
    try:
        global ModulesInstances
        log.debug(f"clean global modules list")
        ModulesInstances = []
    except Exception as ex:
        log.exception(ex)
        log.debug("not ModulesInstances in project")
        pass


# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import inspect
from abc import ABC,abstractmethod
import logging
from logging import Logger as SysLogger
from typing import List

FORMATTER = logging.Formatter("%(levelname)-8s%(filename)s:%(lineno)d  %(message)s")

class Logger:
    __instance = None
    @staticmethod
    def getInstance():
        """ Static access method. """
        if Logger.__instance == None:
            Logger()
        return Logger.__instance

    @staticmethod
    def getLogger() -> logging.Logger:
        return Logger.getInstance().log

    def __init__(self):
        """ Virtually private constructor. """
        if Logger.__instance != None:
            raise Exception("This class is a singleton!")
        else:
            Logger.__instance = self
        self.log = SysLogger.manager.getLogger('pymaketool')
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(FORMATTER)
        self.log.addHandler(console_handler)
        self.log.setLevel(os.environ.get('LOGLEVEL', 'NOTSET'))

__log = Logger().getLogger()

class MKVARS():
    LD      = "$(LD)"
    OBJECTS = '$(OBJECTS)' 
    LDFLAGS = '$(LDFLAGS)'
    OBJCOPY = '$(OBJCOPY)'
    NM          = '$(NM)'
    RANLIB      = '$(RANLIB)'
    STRINGS     = '$(STRINGS)'
    STRIP       = '$(STRIP)'
    CXXFILT     = '$(CXXFILT)'
    ADDR2LINE   = '$(ADD2LINE)'
    READELF     = '$(READELF)'
    CELFEDIT    = '$(ELFEDIT)'
    SIZE    = '$(SIZE)'
    TARGET  = '$(TARGET)'
    PROJECT = '$(PROJECT)'
    STATIC_LIBS = '$(SLIBS_NAMES)'

def MOD_PATH(wk):
    return wk['modPath']


class Define:
    """
    Direct define: { '__USE_FILE__': D('file.h') } => -D__USE_FILE__=file.h
    """
    def __init__(self, value):
        self.value = value
    def getDefine(self):
        if isinstance(self.value, str):
            return self.value
        else:
            return ''
    def __str__(self):
        return str(self.getDefine())
    def __repr__(self):
        return str(self.getDefine())


class AbstractMake(ABC):
    @abstractmethod
    def getProjectSettings(self, **kwargs) -> dict:
        pass
    
    @abstractmethod
    def getTargetsScript(self, **kwargs) -> dict:
        pass

    @abstractmethod
    def getCompilerSet(self, **kwargs) -> dict:
        pass

    @abstractmethod
    def getCompilerOpts(self, **kwargs) -> dict:
        pass

    @abstractmethod
    def getLinkerOpts(self, **kwargs) -> dict:
        pass


def Makeclass(clazz):
    obj = clazz()
    if not isinstance(obj, AbstractMake):
        __log.warning(f"class \'{clazz.__name__}\' in Makefile.py not inheritance of pymakelib.AbstractMake")
    global ProjectInstance
    ProjectInstance = obj


def getProjectInstance() -> AbstractMake:
    try:
        _ = ProjectInstance
        return ProjectInstance
    except NameError:
        __log.debug("not Makeclass mode")
        pass
    return None

## More OOP for pymaketool

from pathlib import Path
import copy
from . import prelib as plib
from . import module
import sys

class Pymaketool():

    def __init__(self, workpath='./'):
        self.workpath = workpath
        sys.path.append(str(self.workpath))
        self.projSettings, self.compilerOpts, self.compilerSettings = plib.read_Makefilepy(self.workpath)


    def getModulesPaths(self) -> list:
        return list(Path(self.workpath).rglob('*[.|_]mk.py'))


    def readModules(self, modulesPaths) -> list:
        self.modules = []
        for filename in modulesPaths:
            mod = plib.readModule(filename, copy.deepcopy(self.compilerOpts), None)
            self.modules.extend(mod)
        return self.modules
    
    def read_modules(self, modulesPaths) -> List[module.AbstractModule]:
        self.modules = []
        for filename in modulesPaths:
            mod = plib.read_module(filename, copy.deepcopy(self.compilerOpts), None)
            self.modules.extend(mod)
        return self.modules


FILE_LANGUAJE_SETTING_XML = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<project>
        <configuration id="0.78511047.1277977235" name="pymaketool">
                <extension point="org.eclipse.cdt.core.LanguageSettingsProvider">
                        <provider copy-of="extension" id="org.eclipse.cdt.ui.UserLanguageSettingsProvider"/>
                        <provider-reference id="org.eclipse.cdt.core.ReferencedProjectsLanguageSettingsProvider" ref="shared-provider"/>
                        <provider-reference id="org.eclipse.cdt.managedbuilder.core.MBSLanguageSettingsProvider" ref="shared-provider"/>
						<!--wildcard_ls_provider-->
                </extension>
        </configuration>
</project>
"""

FILE_PROJECT = """<?xml version="1.0" encoding="UTF-8"?>
<projectDescription>
	<name>{0}</name>
	<comment></comment>
	<projects>
	</projects>
	<buildSpec>
		<buildCommand>
			<name>org.python.pydev.PyDevBuilder</name>
			<arguments>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>org.eclipse.cdt.managedbuilder.core.genmakebuilder</name>
			<triggers>clean,full,incremental,</triggers>
			<arguments>
			</arguments>
		</buildCommand>
		<buildCommand>
			<name>org.eclipse.cdt.managedbuilder.core.ScannerConfigBuilder</name>
			<triggers>full,incremental,</triggers>
			<arguments>
			</arguments>
		</buildCommand>
	</buildSpec>
	<natures>
		<nature>org.eclipse.cdt.core.cnature</nature>
		<nature>org.eclipse.cdt.core.ccnature</nature>
		<nature>org.eclipse.cdt.managedbuilder.core.managedBuildNature</nature>
		<nature>org.eclipse.cdt.managedbuilder.core.ScannerConfigNature</nature>
		<nature>org.python.pydev.pythonNature</nature>
	</natures>
</projectDescription>
"""

FILE_CPROJECT_TEMP = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<?fileVersion 4.0.0?><cproject storage_type_id="org.eclipse.cdt.core.XmlProjectDescriptionStorage">
	<storageModule moduleId="org.eclipse.cdt.core.settings">
		<cconfiguration id="0.78511047.1277977235">
			<storageModule buildSystemId="org.eclipse.cdt.managedbuilder.core.configurationDataProvider" id="0.78511047.1277977235" moduleId="org.eclipse.cdt.core.settings" name="pymaketool">
				<externalSettings/>
				<extensions>
					<extension id="org.eclipse.cdt.core.ELF" point="org.eclipse.cdt.core.BinaryParser"/>
					<extension id="org.eclipse.cdt.core.GASErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GmakeErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GLDErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.VCErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.CWDLocator" point="org.eclipse.cdt.core.ErrorParser"/>
					<extension id="org.eclipse.cdt.core.GCCErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>
				</extensions>
			</storageModule>
			<storageModule moduleId="cdtBuildSystem" version="4.0.0">
				<configuration artifactName="${ProjName}" buildProperties="" description="" id="0.78511047.1277977235" name="pymaketool" optionalBuildProperties="org.eclipse.cdt.docker.launcher.containerbuild.property.volumes=,org.eclipse.cdt.docker.launcher.containerbuild.property.selectedvolumes=" parent="org.eclipse.cdt.build.core.prefbase.cfg">
					<folderInfo id="0.78511047.1277977235." name="/" resourcePath="">
						<toolChain id="org.eclipse.cdt.build.core.prefbase.toolchain.922644828" name="No ToolChain" resourceTypeBasedDiscovery="false" superClass="org.eclipse.cdt.build.core.prefbase.toolchain">
							<targetPlatform binaryParser="org.eclipse.cdt.core.ELF" id="org.eclipse.cdt.build.core.prefbase.toolchain.922644828.1436264033" name=""/>
							<builder id="org.eclipse.cdt.build.core.settings.default.builder.1098600423" keepEnvironmentInBuildfile="false" managedBuildOn="false" name="Gnu Make Builder" superClass="org.eclipse.cdt.build.core.settings.default.builder"/>
							<tool id="org.eclipse.cdt.build.core.settings.holder.libs.236666789" name="holder for library settings" superClass="org.eclipse.cdt.build.core.settings.holder.libs"/>
							<tool id="org.eclipse.cdt.build.core.settings.holder.1768859017" name="Assembly" superClass="org.eclipse.cdt.build.core.settings.holder">
								<inputType id="org.eclipse.cdt.build.core.settings.holder.inType.206126883" languageId="org.eclipse.cdt.core.assembly" languageName="Assembly" sourceContentType="org.eclipse.cdt.core.asmSource" superClass="org.eclipse.cdt.build.core.settings.holder.inType"/>
							</tool>
							<tool id="org.eclipse.cdt.build.core.settings.holder.1825177837" name="GNU C++" superClass="org.eclipse.cdt.build.core.settings.holder">
								<option IS_BUILTIN_EMPTY="false" IS_VALUE_EMPTY="false" id="ilg.gnuarmeclipse.managedbuild.cross.option.cpp.compiler.include.paths.1447306494" name="Include paths (-I)" superClass="ilg.gnuarmeclipse.managedbuild.cross.option.cpp.compiler.include.paths" useByScannerDiscovery="true" valueType="includePath">
									
									<!--wildcard_c_includes-->

								</option>
								<inputType id="org.eclipse.cdt.build.core.settings.holder.inType.1720926587" languageId="org.eclipse.cdt.core.g++" languageName="GNU C++" sourceContentType="org.eclipse.cdt.core.cxxSource,org.eclipse.cdt.core.cxxHeader" superClass="org.eclipse.cdt.build.core.settings.holder.inType"/>
							</tool>
							<tool id="org.eclipse.cdt.build.core.settings.holder.1858450252" name="GNU C" superClass="org.eclipse.cdt.build.core.settings.holder">
								<option IS_BUILTIN_EMPTY="false" IS_VALUE_EMPTY="false" id="org.eclipse.cdt.build.core.settings.holder.incpaths.1905310811" superClass="org.eclipse.cdt.build.core.settings.holder.incpaths" valueType="includePath">
									<!--wildcard_c_includes-->
								</option>
								<option IS_BUILTIN_EMPTY="false" IS_VALUE_EMPTY="false" id="org.eclipse.cdt.build.core.settings.holder.symbols.1546279114" superClass="org.eclipse.cdt.build.core.settings.holder.symbols" valueType="definedSymbols">

									<!--wildcard_c_symbols-->

								</option>
								<inputType id="org.eclipse.cdt.build.core.settings.holder.inType.994335435" languageId="org.eclipse.cdt.core.gcc" languageName="GNU C" sourceContentType="org.eclipse.cdt.core.cSource,org.eclipse.cdt.core.cHeader" superClass="org.eclipse.cdt.build.core.settings.holder.inType"/>
							</tool>
						</toolChain>
					</folderInfo>
					<sourceEntries>
						<!--wildcard_c_exclude-->
					</sourceEntries>
				</configuration>
			</storageModule>
			<storageModule moduleId="org.eclipse.cdt.core.externalSettings"/>
		</cconfiguration>
	</storageModule>
	<storageModule moduleId="cdtBuildSystem" version="4.0.0">
		<project id="stm32h747disco.null.52182215" name="stm32h747disco"/>
	</storageModule>
	<storageModule moduleId="scannerConfiguration">
		<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		<scannerConfigBuildInfo instanceId="ilg.gnuarmeclipse.managedbuild.cross.toolchain.base.463944034;ilg.gnuarmeclipse.managedbuild.cross.toolchain.base.463944034.1798988395;ilg.gnuarmeclipse.managedbuild.cross.tool.c.compiler.2029807856;ilg.gnuarmeclipse.managedbuild.cross.tool.c.compiler.input.2042966102">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
		<scannerConfigBuildInfo instanceId="0.207144966">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
		<scannerConfigBuildInfo instanceId="0.711244623">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
		<scannerConfigBuildInfo instanceId="0.78511047.1277977235">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
		<scannerConfigBuildInfo instanceId="0.207144966;0.207144966.;com.st.stm32cube.ide.mcu.gnu.managedbuild.tool.c.compiler.304600552;com.st.stm32cube.ide.mcu.gnu.managedbuild.tool.c.compiler.input.c.2021731310">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
		<scannerConfigBuildInfo instanceId="0.207144966;0.207144966.;com.st.stm32cube.ide.mcu.gnu.managedbuild.tool.cpp.compiler.1511913342;com.st.stm32cube.ide.mcu.gnu.managedbuild.tool.cpp.compiler.input.cpp.1735336540">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
		<scannerConfigBuildInfo instanceId="ilg.gnuarmeclipse.managedbuild.cross.toolchain.base.463944034;ilg.gnuarmeclipse.managedbuild.cross.toolchain.base.463944034.1798988395;ilg.gnuarmeclipse.managedbuild.cross.tool.cpp.compiler.1603937698;ilg.gnuarmeclipse.managedbuild.cross.tool.cpp.compiler.input.2069329484">
			<autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>
		</scannerConfigBuildInfo>
	</storageModule>
	<storageModule moduleId="org.eclipse.cdt.core.LanguageSettingsProviders"/>
	<storageModule moduleId="refreshScope"/>
	<storageModule moduleId="org.eclipse.cdt.make.core.buildtargets"/>
</cproject>
"""



FILE_MAKEFILE = """# Copyright (c) 2020, Ericson Joseph
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


.DEFAULT_GOAL := all

%: prebuild
\t@time -p $(MAKE) -f makefile.mk $@

prebuild:
\t@pymaketool $(or $(MAKECMDGOALS),all)

.PHONY: test
test_%:
\t@pymaketesting $(subst test_,,$@)
\t$(MAKE) -C Test/ceedling $(subst test_,,$@)

.INTERMEDIATE: prebuild

.PHONY: VERBOSE
ifndef VERBOSE
MAKEFLAGS += --silent
endif
"""

FILE_MAKEFILE_MK = """# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


## Local functions
define logger-compile
	@printf \"%6s\\t%-30s\\n\" $(1) $(2)
endef

define logger-compile-lib
	@printf \"%6s\\t%-25s %-30s\\n\" $(1) $(2) $(3)
endef

.DEFAULT_GOAL := all

CSRC  =
ASSRC = 
INCS  = 
COMPILER_FLAGS =
SLIBS_OBJECTS = 
SLIBS_NAMES = 
SRC_DIRS =

include vars.mk
include srcs.mk

OBJECTS = $(CSRC:%.c=$(PROJECT_OUT)/%.o) $(ASSRC:%.s=$(PROJECT_OUT)/%.o)

include targets.mk

%.o : CFLAGS = $(COMPILER_FLAGS)


$(PROJECT_OUT)/%.o: %.c
	$(call logger-compile,"CC",$<)
	@mkdir -p $(dir $@)
	$(CC) $(CFLAGS) $(INCS) -o $@ -c $<


$(PROJECT_OUT)/%.o: %.s
	$(call logger-compile,"AS",$<)
	@mkdir -p $(dir $@)
	$(CC) $(CFLAGS) $(INCS) -o $@ -c $<


all: $(TARGETS)

clean: clean_targets
\t@echo 'CLEAN'
\trm -f $(addsuffix /*, $(addprefix $(PROJECT_OUT)/,$(SRC_DIRS)))

cleanlibs:
\trm -rf $(SLIBS_OBJECTS:%.a=%.cksum)	

.PHONY: clean cleanlibs

-include $(OBJECTS:.o=.d)

.FORCE:
"""

FILE_MAKEFILE_PY = """import os
from os.path import basename
from pymakelib import git
from pymakelib import MKVARS
from pymakelib.Addon import Addon
from pymakelib.eclipse_addon import EclipseAddon

Addon(EclipseAddon)

def getProjectSettings():
    return {
        'PROJECT_NAME': basename(os.getcwd()),
        'FOLDER_OUT':   'Release/Objects/'
    }


def getTargetsScript():
    PROJECT_NAME = basename(os.getcwd())
    FOLDER_OUT = 'Release/'
    TARGET = FOLDER_OUT + PROJECT_NAME

    TARGETS = {
        'TARGET': {
            'LOGKEY':  'OUT',
            'FILE':    TARGET,
            'SCRIPT':  [MKVARS.LD, '-o', '$@', MKVARS.OBJECTS, MKVARS.LDFLAGS]
        },
        'TARGET_ZIP': {
            'LOGKEY':   'ZIP',
            'FILE':     TARGET + '.zip',
            'SCRIPT':   ['zip', TARGET + '.zip', MKVARS.TARGET]
        }
    }

    return TARGETS


def getCompilerSet():
    return {
        'CC':       'gcc',
        'CXX':      'g++',
        'LD':       'gcc',
        'AR':       'ar',
        'AS':       'as',
        'OBJCOPY':  'objcopy',
        'SIZE':     'size',
        'OBJDUMP':  'objdump',
        'INCLUDES': []
    }


LIBRARIES = []

def getCompilerOpts():

    PROJECT_DEF = {
    }

    return {
        'MACROS': PROJECT_DEF,
        'MACHINE-OPTS': [
        ],
        'OPTIMIZE-OPTS': [
        ],
        'OPTIONS': [
        ],
        'DEBUGGING-OPTS': [
            '-g3'
        ],
        'PREPROCESSOR-OPTS': [
            '-MP',
            '-MMD'
        ],
        'WARNINGS-OPTS': [
        ],
        'CONTROL-C-OPTS': [
            '-std=c89'
        ],
        'GENERAL-OPTS': [
        ],
        'LIBRARIES': LIBRARIES
    }


def getLinkerOpts():
    return {
        'LINKER-SCRIPT': [
        ],
        'MACHINE-OPTS': [
        ],
        'GENERAL-OPTS': [
        ],
        'LINKER-OPTS': [
        ],
        'LIBRARIES': LIBRARIES
    }
"""



# Copyright (c) 2020, Ericson Joseph
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



from typing import List
import sys
from pathlib import Path
import importlib.util
from . import preconts as K
from . import Define as D
from .module import ModuleHandle
from .module import CompilerOptions
from .module import Module, getModuleInstance, cleanModuleInstance, AbstractModule, POJOModule
from .module import StaticLibrary
from . import getProjectInstance
from . import Logger

log = Logger.getLogger()

def add_value2list(dstList: list, values):
    if isinstance(values, list):
        for item in values:
            dstList.append(item)
    elif isinstance(values, dict):
        for keys in values:
            if isinstance(values[keys], list):
                for item in values[keys]:
                    dstList.append(item)
            else:
                dstList.append(values[keys])
    else:
        dstList.append(values)

def wprInit(mod, modHandle, moduleInstance=None):
    try:
        if moduleInstance:
            return moduleInstance.init()
        else:
            return getattr(mod, K.MOD_F_INIT)(modHandle)
    except AttributeError as ae:
        log.debug(ae)
    except Exception as ex:
        log.exception(ex)
        exit(-1)

def wprGetSrcs(mod, modHandle, moduleInstance=None):
    try:
        if moduleInstance:
            return moduleInstance.getSrcs()
        else:
            return getattr(mod, K.MOD_F_GETSRCS)(modHandle)
    except AttributeError as ae:
        log.warning(ae)
    except Exception as ex:
        log.exception(ex)
        exit(-1)

def wprGetIncs(mod, modHandle, moduleInstance=None):
    try:
        if moduleInstance:
            return moduleInstance.getIncs()
        else:
            return getattr(mod, K.MOD_F_GETINCS)(modHandle)
    except AttributeError as ae:
        log.warning(ae)
    except Exception as ex:
        log.exception(ex)
        exit(-1)

def wprGetCompilerOpts(mod, modHandle, moduleInstance=None):
    try:
        if moduleInstance:
            return moduleInstance.getCompilerOpts()
        else:
            return getattr(mod, K.MOD_F_GETCOMPILEROPTS)(modHandle)
    except AttributeError as ae:
        log.debug(ae)
    except Exception as ex:
        log.exception(ex)
        exit(-1)

def reafGenHeader(headerpath):
    lib = importlib.util.spec_from_file_location(str(headerpath), str(headerpath))
    mod = importlib.util.module_from_spec(lib)
    log.debug(f"exec code generator {mod.__name__}")
    outfile = str(headerpath)
    outfile = outfile.replace('_h.py', '.h')
    outfile = outfile.replace('.h.py', '.h')
    log.debug(f"output fiel {outfile}")
    stdout_ = sys.stdout #Keep track of the previous value.
    sys.stdout = open(outfile, 'w') # Something here that provides a write method.
    lib.loader.exec_module(mod)
    sys.stdout = stdout_


def read_module(module_path: Path, compiler_opts, goals=None) -> List[AbstractModule]:
    lib = importlib.util.spec_from_file_location(str(module_path), str(module_path))
    mod = importlib.util.module_from_spec(lib)
    log.debug(f"exec module {mod.__name__}")
    lib.loader.exec_module(mod)
    
    moduleInstances = getModuleInstance()
    modules = []

    if not moduleInstances:
        modHandle = ModuleHandle(module_path.parent, compiler_opts, goals)
        m = POJOModule(module_path)
        m.init_resp = wprInit(mod, modHandle)
        m.includes = wprGetIncs(mod, modHandle)
        m.sources = wprGetSrcs(mod, modHandle)
        m.compiler_opts = wprGetCompilerOpts(mod, modHandle)
        modules.append(m)
    else:
        modules.extend(getModuleInstance())

    cleanModuleInstance()
    
    return modules


def readModule(modPath, compilerOpts, goals=None):
    lib = importlib.util.spec_from_file_location(str(modPath), str(modPath))
    mod = importlib.util.module_from_spec(lib)
    log.debug(f"exec module {mod.__name__}")
    lib.loader.exec_module(mod)

    moduleInstances = getModuleInstance()
    modules = []

    if not moduleInstances:
        modHandle = ModuleHandle(modPath.parent, compilerOpts, goals)
        srcs = []
        incs = []
        flags = []
        staticLib = None
        # init attribute is not mandatory
        result = wprInit(mod, modHandle)
        if isinstance(result, StaticLibrary):
            log.debug(f"\'{mod.__name__}\' is static library")
            staticLib = result

        result = wprGetSrcs(mod, modHandle)
        if result:
            add_value2list(srcs, result)
        else:
            log.debug(f"\'{mod.__name__}\' return empty sources")

        result = wprGetIncs(mod, modHandle)
        if result:
            add_value2list(incs, result)
        else:
            log.debug(f"\'{mod.__name__}\' return empty includes")

        result = wprGetCompilerOpts(mod, modHandle)
        if issubclass(type(result), CompilerOptions):
            result = result.opts
        if result:
            flags.append(result)
        else:
            log.debug(
                f"\'{mod.__name__}\' return empty compiler options")

        modules.append(Module(srcs, incs, flags, modPath, staticLib=staticLib))
        return modules

    for moduleInstance in moduleInstances:

        modHandle = ModuleHandle(modPath.parent, compilerOpts, goals)

        srcs = []
        incs = []
        flags = []
        staticLib = None

        if moduleInstance:
            log.info(f"read ModuleClass \'{mod.__name__}:{type(moduleInstance).__name__}\'")

        # init attribute is not mandatory
        result = wprInit(mod, modHandle, moduleInstance)
        if isinstance(result, StaticLibrary):
            log.debug(f"\'{type(moduleInstance).__name__}\' is static library")
            staticLib = result

        result = wprGetSrcs(mod, modHandle, moduleInstance)
        if result:
            add_value2list(srcs, result)
        else:
            log.debug(f"\'{type(moduleInstance).__name__}\' return empty sources")

        result = wprGetIncs(mod, modHandle, moduleInstance)
        if result:
            add_value2list(incs, result)
        else:
            log.debug(f"\'{type(moduleInstance).__name__}\' return empty includes")

        result = wprGetCompilerOpts(mod, modHandle, moduleInstance)
        if issubclass(type(result), CompilerOptions):
            result = result.opts
        if result:
            flags.append(result)
        else:
            log.debug(
                f"\'{type(moduleInstance).__name__}\' return empty compiler options")

        m = Module(srcs, incs, flags, modPath, staticLib=staticLib)
        if moduleInstance.module_name:
            m.module_name = moduleInstance.module_name
        modules.append(m)

    cleanModuleInstance()

    return modules


def list2str(l):
    return ' '.join(l)


def macrosDictToString(macros):
    mstr = []
    if isinstance(macros, dict):
        for key in macros:
            if macros[key] != None and macros[key] != '':
                if isinstance(macros[key], str):
                    mstr.append('-D{}=\\\"{}\\\"'.format(key, macros[key]))
                elif isinstance(macros[key], bool):
                    mstr.append(
                        '-D{}={}'.format(key, '1' if macros[key] else '0'))
                elif isinstance(macros[key], D):
                    mstr.append(
                        '-D{}={}'.format(key, macros[key].getDefine()))
                else:
                    mstr.append('-D{}={}'.format(key, macros[key]))
            else:
                mstr.append('-D{}'.format(key))

    return ' '.join(mstr)


def compilerOptsByModuleToLine(compOpts):
    mstr = []
    for moduleCompileOps in compOpts:
        if isinstance(moduleCompileOps, dict):
            for key in moduleCompileOps:
                if key == 'TARGETS':
                    continue
                if (key == K.COMPOPTS_MACROS_KEY and isinstance(moduleCompileOps[key], dict)):
                    macros = macrosDictToString(moduleCompileOps[key])
                    mstr.append(macros)
                else:
                    mstr.append(list2str(moduleCompileOps[key]))

        elif isinstance(moduleCompileOps, list):
            for item in moduleCompileOps:
                mstr.append(item)
    rmstr = list(filter(lambda item: item, mstr))
    rmstr = ' '.join(rmstr)
    rmstr = ' '.join(rmstr.split())
    log.debug(f"compiler options: {rmstr}")
    return rmstr


def read_Makefilepy(workpath=''):
    makefilepy_path = workpath + K.MAKEFILE_PY
    lib = importlib.util.spec_from_file_location(makefilepy_path, makefilepy_path)
    mod = importlib.util.module_from_spec(lib)
    lib.loader.exec_module(mod)

    projectInstance = getProjectInstance()
    if projectInstance:
        log.info("makeclass define")

    def wprGetProjectSettings():
        try:
            if projectInstance:
                return projectInstance.getProjectSettings()
            else:
                return getattr(mod, K.MK_F_GETPROJECTSETTINGS)()
        except Exception as ex:
            log.exception(ex)
            exit(-1)

    def wprGetCompilerSet():
        try:
            if projectInstance:
                return projectInstance.getCompilerSet()
            else:
                return getattr(mod, K.MK_F_GETCOMPILERSET)()
        except Exception as ex:
            log.exception(ex)
            exit(-1)

    def wprGetCompileOpts():
        try:
            if projectInstance:
                return projectInstance.getCompilerOpts()
            else:
                return getattr(mod, K.MK_F_GETCOMPILEROPTS)()
        except Exception as ex:
            log.exception(ex)
            exit(-1)

    def wprGetLinkerOpts():
        try:
            if projectInstance:
                return projectInstance.getLinkerOpts()
            else:
                return getattr(mod, K.MK_F_GETLINKEROPTS)()
        except Exception as ex:
            log.exception(ex)
            exit(-1)

    def wprGetTargetScript():
        try:
            if projectInstance:
                return projectInstance.getTargetsScript()
            else:
                return getattr(mod, K.MK_F_GETTARGETSSCRIPT)()
        except Exception as ex:
            log.exception(ex)
            exit(-1)

    makevars = open(K.VARS_MK, 'w')

    projSettings = None
    compSet = None
    compOpts = None

    try:
        projSettings = wprGetProjectSettings()
        if projSettings[K.PROJSETT_PROJECTNAME]:
            makevars.write('{0:<15} = {1}\n'.format(
                'PROJECT', projSettings[K.PROJSETT_PROJECTNAME]))
        if projSettings[K.PROJSETT_FOLDEROUT]:
            projSettings[K.PROJSETT_FOLDEROUT] = Path(
                projSettings[K.PROJSETT_FOLDEROUT])
            makevars.write('{0:<15} = {1}\n'.format(
                'PROJECT_OUT', projSettings[K.PROJSETT_FOLDEROUT]))
    except Exception as e:
        log.exception(e)

    makevars.write('\n')

    try:
        compSet = wprGetCompilerSet()
        for sfx in (
            K.COMPILERSET_CC,
            K.COMPILERSET_CXX,
            K.COMPILERSET_LD,
            K.COMPILERSET_AR,
            K.COMPILERSET_AS,
            K.COMPILERSET_OBJCOPY,
            K.COMPILERSET_SIZE,
            K.COMPILERSET_OBJDUMP,
            K.COMPILERSET_NM,
            K.COMPILERSET_RANLIB,
            K.COMPILERSET_STRINGS,
            K.COMPILERSET_STRIP,
            K.COMPILERSET_CXXFILT,
            K.COMPILERSET_ADDR2LINE,
            K.COMPILERSET_READELF,
            K.COMPILERSET_ELFEDIT
        ):
            if compSet[sfx]:
                makevars.write('{0:<10} := {1}\n'.format(sfx, compSet[sfx]))
    except KeyError as ke:
        log.debug(f"not found compiler option {ke}")
    except Exception as e:
        log.exception(e)

    makevars.write('\n')

    try:
        compOpts = wprGetCompileOpts()
        if isinstance(compOpts, dict):
            for key in compOpts:
                makevars.write('# {0}\n'.format(key))
                if (key == 'MACROS' and isinstance(compOpts[key], dict)):
                    makevars.write(
                        'COMPILER_FLAGS += {}\n'.format(macrosDictToString(compOpts[key])))
                else:
                    makevars.write(
                        'COMPILER_FLAGS += {}\n'.format(list2str(compOpts[key])))
        elif isinstance(compOpts, list):
            for item in compOpts:
                makevars.write('COMPILER_FLAGS += {}\n'.format(item))
        else:
            log.debug("Not load getCompilerOpts")
    except Exception as ex:
        log.exception(ex)

    makevars.write('\n')

    try:
        linkOpts = wprGetLinkerOpts()
        if isinstance(linkOpts, dict):
            for keys in linkOpts:
                makevars.write('# {0}\n'.format(keys))
                makevars.write(
                    'LDFLAGS += {}\n'.format(list2str(linkOpts[keys])))
        elif isinstance(linkOpts, list):
            for item in linkOpts:
                makevars.write('LDFLAGS += {}\n'.format(item))
    except Exception as ex:
        log.exception(ex)

    makevars.close()

    targetsmk = open('targets.mk', 'w')

    try:
        targets = wprGetTargetScript()
        if isinstance(targets, dict):
            if len(targets) == 0:
                pass
            else:
                labels = []
                targetval = []
                targetsct = []
                logkeys = []
                for k in targets:
                    labels.append(k)
                    targetval.append(targets[k]['FILE'])
                    targetsct.append(targets[k]['SCRIPT'])
                    if 'LOGKEY' in targets[k]:
                        logkeys.append(targets[k]['LOGKEY'])
                    else:
                        logkeys.append('>>')

                # print labels
                for i in range(len(targetval)):
                    targetsmk.write("{0:<15} = {1}\n".format(
                        labels[i], targetval[i]))

                targetsmk.write('\nTARGETS = $({})\n\n'.format(
                    labels[len(targetval)-1]))

                for i in range(len(labels)):
                    if labels[i] == 'TARGET':
                        targetsmk.write("\n$({}): {} {}\n".format(
                            labels[i], '$(OBJECTS) $(SLIBS_OBJECTS)', '' if i == 0 else "$("+ labels[i-1] + ")"))
                    else:
                        targetsmk.write("\n$({}): {}\n".format(
                            labels[i], '' if i == 0 else "$("+ labels[i-1] + ")"))

                    targetsmk.write(
                        '\t$(call logger-compile,"{}",$@)\n'.format(logkeys[i]))
                    script = targetsct[i]
                    script = ' '.join(script)
                    targetsmk.write('\t{}\n'.format(script))

                targetsmk.write('\n')

                targetsmk.write("\n{}:\n".format('clean_targets'))
                targetlist = ('$('+l+')' for l in labels)
                targetsmk.write('\trm -rf {}\n'.format(' '.join(targetlist)))
                if compOpts:
                    compOpts['TARGETS'] = targets
    except Exception as e:
        log.exception(e)

    targetsmk.close()

    return projSettings, compOpts, compSet



# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from pathlib import Path
import subprocess
from . import moduleignore

def getCommitHash(abbreviated=True):
    hashValue = ''
    cmd = ['git', 'rev-parse', 'HEAD']
    try:
        hashValue = subprocess.check_output(cmd).strip().decode('utf-8')
        if hashValue:
            hashValue = hashValue[:7] if abbreviated else hashValue
    except:
        print('WARNING: Can not get commit hash')
        
    return hashValue


def getBranchName():
    branchName = ''
    try:
        branchName = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip().decode('utf-8')
    except:
        print('WARNING: Can not get commit hash')

    return branchName


def getDescribe(options='--long'):
    desc = ''
    try:
        desc = subprocess.check_output(['git','describe', options]).strip().decode('utf-8')
    except:
        print('WARNING: Can not get commit hash')
        
    return desc


def printRelativePath(filemacro):
    print(os.path.dirname(os.path.abspath(filemacro)))


@DeprecationWarning
def addSubmodule(url, dst_dir, isPyMakeModule=False, exclModulesPaths=[]):
    dstDir = Path(Path(dst_dir) / '.git')
    if dstDir.exists():
        moduleignore.writeIgnoreFile(exclModulesPaths)
        return

    cmd1 = ['git',  'submodule', 'add', url, dst_dir]
    cmd2 = ['git',  'submodule', 'init']
    cmd3 = ['git',  'submodule', 'update']
    
    try:
        subprocess.check_output(cmd1).strip().decode('utf-8')
    except:
        pass

    try:
        subprocess.check_output(cmd2).strip().decode('utf-8')
    except:
        pass

    try:
        subprocess.check_output(cmd3).strip().decode('utf-8')
    except:
        pass

    if isPyMakeModule:
        moduleignore.writeIgnoreFile(exclModulesPaths)



# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import re
import sys
import getopt


class ItemSizeStat:
    def __init__(self, name, size, addr):
        self.name = name
        self.size = size
        self.addr = addr

    def string(self):
        return self.name + " " + self.size + " " + self.addr

    def getAddr(self):
        return self.addr

    def getSize(self):
        return self.size


def getSizeAddr(line):
    x = re.search("^([A-Za-z0-9_\.]+)[ ]+([a-z0-9]+)[ ]+([a-z0-9]+).*", line)
    if(x):
        return ItemSizeStat(x.group(1), x.group(2), x.group(3))
    else:
        None


def printKB(value, decimals=1):
    return ('{:.'+str(decimals)+'f}').format(value/1000)

def printPrtg(value, decimals=1):
    return ('{:.'+str(decimals)+'f}').format(value*100)


def main(argv):
    totalFlashLen = 0
    totalRAMLen = 0
    sizefile = ''

    try:
        opts, args = getopt.getopt(
            argv, "hF:R:s:", ["flash-len=", "ram-len=", "size-file="])
    except getopt.GetoptError:
        print('armsize.py -F <flash-len:int> -R <ram-len:int> -s <size-file:file>')
        sys.exit(2)

    for opt, arg in opts:
        if opt in ('-F', '--flash-len'):
            totalFlashLen = int(arg)
        elif opt in ('-R', '--ram-len'):
            totalRAMLen = int(arg)
        elif opt in ('-s', '--size-file'):
            sizefile = arg

    file = open(sizefile, "r")

    STATE_FIND = 0
    STATE_READ = 1
    STATE_END = 2

    state = STATE_FIND

    itemsSize = []

    for line in file:
        if(state == STATE_FIND):
            x = re.search("^section[ ]+size[ ]+addr.*", line)
            if(x):
                state = STATE_READ
        elif state == STATE_READ:
            x = re.search("^Total.*", line)
            if(x):
                state = STATE_END
            else:
                item = getSizeAddr(line)
                itemsSize.append(item)

    RAM_ADDR = int("0x20000000", 16)
    FLASH_ADDR = int("0x8000000", 16)
    TOTAL_RAM = float(totalRAMLen)
    TOTAL_FLASH = float(totalFlashLen)
    RAM_USAGE = 0
    FLASH_USAGE = 0

    for item in itemsSize:
        itemAddr = int(item.getAddr(), 16)
        if(itemAddr >= RAM_ADDR):
            RAM_USAGE = RAM_USAGE + int(item.getSize(), 16)
        elif(itemAddr >= FLASH_ADDR and itemAddr < RAM_ADDR):
            FLASH_USAGE = FLASH_USAGE + int(item.getSize(), 16)

    print('{:<6}{:>14}{:>14}{:>14}'.format('REGION', 'SIZE', 'USED', 'USAGE (%)'))

    perc = (float(RAM_USAGE) / TOTAL_RAM)
    print("RAM:   {:>10} KB {:>10} KB {:>10} %".format(
        printKB(TOTAL_RAM), printKB(RAM_USAGE), printPrtg(perc)))

    perc = (float(FLASH_USAGE) / TOTAL_FLASH)
    print("FLASH: {:>10} KB {:>10} KB {:>10} %".format(
        printKB(TOTAL_FLASH), printKB(FLASH_USAGE), printPrtg(perc)))


if __name__ == "__main__":
    main(sys.argv[1:])



from pathlib import Path
from . import preconts

def readIgnoreFile(file = Path(preconts.MODULEIGNORE_FILE)):
    ignorelist = []
    try:
        ignoreFile = open(str(file), 'r')
        for line in ignoreFile:
            ignorelist.append(Path(line.rstrip()))
    except:
        pass
    return ignorelist


def writeIgnoreFile( ignoreList: list, file = Path(preconts.MODULEIGNORE_FILE)):

    currList = readIgnoreFile(file)

    try:

        ignoreFile = open(str(file), "w")

        for item in ignoreList:
            if not Path(item) in currList:
                currList.append(item)
        
        ignores = list(map(str, currList))
        ignores = map(lambda x: x + '\n', ignores)
        ignoreFile.writelines(ignores)

    except Exception as e:
        print(e)


# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from pathlib import Path
from .module import ModuleHandle
from .module import SrcType
from .module import IncType

def getAllSrcs(wkmh, srcType: SrcType):
    if isinstance(wkmh, ModuleHandle):
        wk = wkmh.getWorkspace()
    else:
        wk = wkmh
    srcs = []
    for ext in srcType:
        srcs += list(Path(wk['modPath']).rglob('*' + ext))
    return srcs


def getAllSrcs_C(wkmh):
    if isinstance(wkmh, ModuleHandle):
        wk = wkmh.getWorkspace()
    else:
        wk = wkmh
    return getAllSrcs(wk, SrcType.C)


def getSrcsByRgx(wkmh, *regexs):
    if isinstance(wkmh, ModuleHandle):
        wk = wkmh.getWorkspace()
    else:
        wk = wkmh
    srcs = []
    for r in regexs:
        srcs += list(Path(wk['modPath']).rglob(r))

    srcs = list(dict.fromkeys(srcs))
    return srcs


def getAllIncs(wkmh, incType: IncType):
    if isinstance(wkmh, ModuleHandle):
        wk = wkmh.getWorkspace()
    else:
        wk = wkmh
    incsfiles = []
    for ext in incType:
        incsfiles += list(Path(wk['modPath']).rglob('*' + ext))

    incs = []
    for i in incsfiles:
        incs.append(i.parent)

    incs = list(dict.fromkeys(incs))
    return incs


def getAllIncs_C(wkmh):
    if isinstance(wkmh, ModuleHandle):
        wk = wkmh.getWorkspace()
    else:
        wk = wkmh
    return getAllIncs(wk, IncType.C)



# Copyright (c) 2021, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pymakelib import Define as D

def out(value):
    lines = value.splitlines()
    for l in lines:
        print(l)

def comment(value: str):
    lines = value.splitlines()
    out("/**")
    for l in lines:
        out(" * "+l.strip())
    out(" */")

def enum(names, values=[0]):
    resp = ''
    resp += "enum {\n"
    idx = 0
    for n in names:
        v = None
        try:
            v = values[idx]
        except:
            v = None
            pass
        if isinstance(v, int):
            resp += f"    {n} = {v},\n"
        else:
            resp += f"    {n},\n"
        idx += 1
    resp += "};"
    out(resp)


def enum_sf(strformat, range, init=0):
    values = []
    for r in range:
        values.append(strformat.format(r))
    enum(values, [init])

def enum_str_map(name, strdict: dict):
    keys = list(strdict.keys())
    enum(keys)
    out("""#define {0}_VALUES  =\\
    {{\\
{1}\\
    }}
    """.format(name.upper(),'\\\n'.join([ f"{' ':8}\"{strdict[n]}\"," for n in keys])))
    out("""
#ifndef DECL_{0}
#define _DECL extern
#define _VAR
#else
#define _DECL
#define _VAR  {0}_VALUES
#endif
    """.format(name.upper()))
    out("""
_DECL char* {0}[{1}] _VAR;
    """.format(name, len(keys)))

def __format_name(name:str) -> str:
    name = name.upper()
    name = name.replace('/', '_')
    name = name.replace('.', '_')
    if not name.startswith('_'):
        name = '_'+name
    if not name.endswith('_'):
        name = name + '_'
    return name 

def HEADER_FILE(*args, **kwargs):
    
    def inner(func):
        if 'name' in kwargs.keys():
            name =  kwargs['name']
        else:
            name = func.__name__
        name = __format_name(name)
        out(f"#ifndef {name}")
        out(f"#define {name}")
        func()
        out(f"#endif // {name}")
         
    # reurning inner function    
    return inner




# Copyright (c) 2021, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pymakelib import Pymaketool
from pymakelib import Logger
from pymakelib.module import CompilerOptions
from pymakelib import Define

log = Logger.getLogger()

def setSettings(settings):
    global __settings
    try:
        _ = __settings
    except NameError:
        log.debug(f"create global modules list")
        __settings = settings


def getSettings():
    auxsetts = None
    try:
        _ = __settings
        return __settings
    except NameError:
        log.debug("not project settings found")
        p = Pymaketool()
        globalSettings = {
            'PROJECT_SETTINGS':  p.projSettings,
            'COMPILER_OTPS':     p.compilerOpts,
            'COMPILER_SETTINGS': p.compilerSettings
        }
        setSettings(globalSettings)
        auxsetts = __settings
        pass
    return auxsetts

def isdefined(key) -> bool:
    """Check if project have define.

    Args:
        key (str or D): name of define or macro

    Returns:
        bool: True if key is defined
    """
    sett = getSettings()
    macros = sett['COMPILER_OTPS']['MACROS']
    return (key in macros)


def define(key) -> str:
    """Get value of define if exist.

    Args:
        key (str): name of define (macro)

    Returns:
        str: value of define in string, if define value is None return '',if key is not defined return None
    """    
    sett = getSettings()
    macros = sett['COMPILER_OTPS']['MACROS']
    if isdefined(key):
        value = macros[key]
        return '' if not value else str(value)
    else:
        return None


def getCompilerOpts() ->  dict:
    """Get the project compiler options

    Returns:
        dict: General project compiler options
    """
    return getSettings()['COMPILER_OTPS']



# Copyright (c) 2020, Ericson Joseph
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of pyMakeTool nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sys


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    CEND = '\33[0m'
    CBOLD = '\33[1m'
    CITALIC = '\33[3m'
    CURL = '\33[4m'
    CBLINK = '\33[5m'
    CBLINK2 = '\33[6m'
    CSELECTED = '\33[7m'
    CBLACK = '\33[30m'
    CRED = '\33[31m'
    CGREEN = '\33[32m'
    CYELLOW = '\33[33m'
    CBLUE = '\33[34m'
    CVIOLET = '\33[35m'
    CBEIGE = '\33[36m'
    CWHITE = '\33[37m'
    CBLACKBG = '\33[40m'
    CREDBG = '\33[41m'
    CGREENBG = '\33[42m'
    CYELLOWBG = '\33[43m'
    CBLUEBG = '\33[44m'
    CVIOLETBG = '\33[45m'
    CBEIGEBG = '\33[46m'
    CWHITEBG = '\33[47m'
    CGREY = '\33[90m'
    CRED2 = '\33[91m'
    CGREEN2 = '\33[92m'
    CYELLOW2 = '\33[93m'
    CBLUE2 = '\33[94m'
    CVIOLET2 = '\33[95m'
    CBEIGE2 = '\33[96m'
    CWHITE2 = '\33[97m'
    CGREYBG = '\33[100m'
    CREDBG2 = '\33[101m'
    CGREENBG2 = '\33[102m'
    CYELLOWBG2 = '\33[103m'
    CBLUEBG2 = '\33[104m'
    CVIOLETBG2 = '\33[105m'
    CBEIGEBG2 = '\33[106m'
    CWHITEBG2 = '\33[107m'


def colorSrc(pathsrc):
    strpath = str(pathsrc)
    aux = strpath.split('/')
    src = aux[-1]
    parent = '/'.join(aux[:-1])
    return '{0}{2}/{1}{3}'.format(bcolors.CGREY, bcolors.ENDC, parent, src)


if (len(sys.argv) > 2):
    if sys.argv[1] == 'CC':
        print('{0}{2:>6}{1}\t{3}'.format(bcolors.OKGREEN,
                                         bcolors.ENDC, 'CC', colorSrc(sys.argv[2])))
    elif sys.argv[1] == 'AS':
        print('{0}{2:>6}{1}\t{3}'.format(bcolors.OKBLUE,
                                         bcolors.ENDC, 'AS', colorSrc(sys.argv[2])))
    elif sys.argv[1] == 'LD':
        print('{0}{2:>6}{1}\t{3}'.format(bcolors.HEADER,
                                         bcolors.ENDC, 'LD', colorSrc(sys.argv[2])))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pymaketool",
    version="2.0.3",
    author="Ericson Joseph",
    author_email="ericsonjoseph@gmail.com",
    description="Python Makefile Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    scripts = ['scripts/pymaketool', 'scripts/pymaketesting', 'scripts/pybuildanalyzer', 'scripts/pybuildanalyzer2', 'scripts/pymakedot', 'scripts/pynewproject'],
    url="https://github.com/ericsonj/pymaketool",
    license="MIT",
    packages=setuptools.find_packages(),
    package_data={'': ['*.glade', '*.tar.gz']},
    install_requires=[
        'prompt_toolkit',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import datetime
import os
from pathlib import Path
from pymakelib import nproject
from pkg_resources import resource_filename
import argparse

class CLinuxGCC(nproject.BasicGenerator):
    """Generate basic C linux gcc project 
    """
    def info(self):
        return {
            "name": "C Project",
            "desc": "C Linux GCC Project"
        }

    def print_help(self):
        print("""
Basic C Linux GCC Project

Tokens:
    author
    project_name

Command: pynewproject CLinuxGCC author=\\\"Author\\\" project_name=project
        """)

    def get_temp_files(self):
        return [
            'app/inc/main.h',
            'app/src/main.c',
            'app/app_mk.py',
            'makefile.mk',
            'Makefile',
            'Makefile.py',
            ".project",
            ".pymakeproj/.cproject_template",
            ".pymakeproj/.language.settings_template",
            ".settings",
        ]

    def get_attrs(self, **kwargs) -> dict:

        args = self.parse_args(kwargs['args'])

        if 'help' in args:
            self.print_help()
            exit(0)

        tokens = {
            'author':       'Ericson Joseph',
            'date':         datetime.datetime.now().strftime("%b %d %Y"),
            'project_name': 'c_temp'
        }
        tokens['author']       = input("(author) Your name: ") if not 'author' in args else args['author']
        tokens['project_name'] = input("(project_name) Your project name: ") if not 'project_name' in args else args['project_name']

        output_dir = Path( Path(os.getcwd()) / Path(tokens['project_name'])) 

        gzip_file = resource_filename("pymakelib.resources.templates", "clinuxgcc.tar.gz")

        return {
            "temp_name":        "clinuxgcc",
            "temp_gzip_file":   gzip_file,
            "temp_files":       self.get_temp_files(),
            "temp_tokens":      tokens,
            "output_folder":    output_dir,
        }



import subprocess
import tempfile


def getCeedlingHeaderFiles():
    ceedling_includes = []
    try:
        auxfile = tempfile.NamedTemporaryFile()
        command = "find /var/lib/gems $HOME/.gem/ -name 'unity.h' 2>&1 | grep '.*ceedling-[0-9\\.\\-]*/vendor/unity/src'"
        res = subprocess.check_output(["bash", "-c", command])
        for line in res.splitlines():
            ceedling_includes.append(line.decode('utf-8').strip())
        auxfile.close()
    except Exception as e:
        print(e)
    return ceedling_includes